%
% (c) Andre Santos
%
\section{Generate CSharp}

Name mangling for CSharp.
~~~~~~~~~~~~~~~~~~~~~~

Haskell has a number of namespaces. The CSharp translator uses
the standard Haskell mangles (see OccName.lhs), and some extra
mangles.

All names are hidden inside packages.

module name:
  - becomes a first level csharp package.
  - can not clash with csharp, because haskell modules are upper case,
     csharp default packages are lower case.

function names:
  - these turn into classes
  - csharp keywords (eg. private) have the suffix "zdk" ($k) added.

data *types*
  - These have a base class, so need to appear in the
    same name space as other object. for example data Foo = Foo
  - We add a postfix to types: "zdc" ($c)
  - Types are upper case, so never clash with keywords

data constructors
  - There are tWO classes for each Constructor
   (1) - Class with the payload extends the relevent datatype baseclass.
       - This class has the prefix zdw ($w)
   (2) - Constructor *wrapper* just use their own name.
    - Constructors are upper case, so never clash with keywords
    - So Foo would become 2 classes.
	* Foo		-- the constructor wrapper
	* zdwFoo	-- the worker, with the payload


$i  for instances.
$k  for keyword nameclash avoidance.

\begin{code}
module CSharpGen ( csharpGen ) where

import CSharp

import Literal	( Literal(..) )
import Id	( Id, isDataConWorkId_maybe, isId, idName,
                  idType, isDeadBinder, idPrimRep
		, isPrimOpId_maybe )
import Name	( NamedThing(..), getOccString, isExternalName, isInternalName
		, nameModule )
import PrimRep  ( PrimRep(..) )
import DataCon	( DataCon, dataConRepArity, dataConRepArgTys, dataConWorkId )
import qualified Type
import qualified CoreSyn
import StgSyn
import CoreSyn	( CoreBind, CoreExpr, CoreAlt, CoreBndr,
		  Bind(..), AltCon(..)
		)
import TysWiredIn	( boolTy, trueDataCon, falseDataCon )
import qualified CoreUtils
import Module	( Module, moduleString )
import TyCon	( TyCon, isDataTyCon, tyConDataCons )
import Outputable

import Maybe
import PrimOp
import Util     ( lengthIs, notNull )
import FastString
import Debug.Trace (trace)

#include "HsVersions.h"

\end{code}


\begin{code}
csharpGen :: Module -> [TyCon] -> [(StgBinding,[(Id,[Id])])] -> CompilationUnit

csharpGen mod tycons binds
  = {- liftCompilationUnit -} package
  where
    decls = [Import "HaskellRTS"] ++
             -- [Import (moduleString mod) | mod <- import_mods] ++
             concat (map csharpTyCon (filter isDataTyCon tycons)) 
             ++ mainClass (concat  $ map csharpTopBind binds)
    package = Package (module_name) decls
    mainClass decls  
	= [ Class [Public] (module_name) [] [] decls ]
    module_name = moduleString mod
\end{code}


%************************************************************************
%*									*
\subsection{Type declarations}
%*									*
%************************************************************************

\begin{code}
csharpTyCon :: TyCon -> [Decl]
--  	public class List {}
--
--	public class $wCons extends List {
--		Object f1; Object f2
--	}
--	public class $wNil extends List {}

-- ToDo: support partial application of constructors, through functions

csharpTyCon tycon
  = tycon_jclass : concat (map constr_class constrs)
  where
    constrs = tyConDataCons tycon
    tycon_jclass_jname =  csharpTyConTypeName tycon 
    tycon_jclass = Class [Public] (shortName tycon_jclass_jname) [closureName] [] []

    constr_class data_con
	= [ Class [Public] constr_jname [tycon_jclass_jname] []
				(field_decls ++ [cons_meth])
	  ]
	where
	  constr_jname = shortName (csharpConstrWkrName data_con)

	  field_names  = constrToFields data_con
	  field_decls  = [ Field [Public] n Nothing 
			 | n <- field_names
			 ]

	  cons_meth    = mkCons constr_jname field_names


-- This checks to see if the type is reasonable to call new with.
-- primitives might use a static method later.
mkNew :: Type -> [Expr] -> Expr
mkNew t@(PrimType primType) _  = error "new primitive -- fix it???"
mkNew t@(Type _)            es = New t es Nothing
mkNew _                     _  = error "new with strange arguments"

constrToFields :: DataCon -> [Name]
constrToFields cons = 
	[ fieldName i t 
	| (i,t) <- zip [1..] (map primRepToType
			          (map Type.typePrimRep
				       (dataConRepArgTys cons)
				  )
			     )
	]

mkCons :: TypeName -> [Name] -> Decl
mkCons name args = Constructor [Public] name
	[ Parameter [] n | n <- args ]
	[ ExprStatement (Assign 
			   (Access this n)
			   (Var n)
			 )
		    | n <- args ]

-- not used:
-- mkStr :: String -> Expr
-- mkStr str = Literal (StringLit str)
\end{code}

%************************************************************************
%*									*
\subsection{Bindings}
%*									*
%************************************************************************

\begin{code}
csharpTopBind :: (StgBinding,[(Id,[Id])]) -> [Decl]
csharpTopBind (StgNonRec bndr rhs,_) = csharp_top_bind (bndr,rhs)
csharpTopBind (StgRec prs,_) 	      = concat $ map csharp_top_bind prs

csharp_top_bind :: (Id, StgRhs) -> [Decl]
-- 	public class f implements Code {
--	  public Object ENTER() { ...translation of rhs... }
--	}

csharp_top_bind (bndr, StgRhsCon _cc con args)
  = [Method [Public] (csharpName bndr) [] []
            [Return (New (csharpConstrWkrType con) (csharpArgs args) Nothing) ]
    ]

-- no args or fvs
csharp_top_bind (bndr, StgRhsClosure _ _ [] upd _ [] rhs)
  = [Method [Public] (csharpName bndr) [] []
            (csharpStgExpr Return rhs)
    ]

-- no fvs (e.g. top level)
csharp_top_bind (bndr, StgRhsClosure _ _ [] upd _ args rhs)
  = [Method [Public] ((csharpName bndr))
            (map argName args)
            []
            (csharpStgExpr Return rhs)
    ]
  where
    argName :: Id -> Parameter
    argName id = Parameter [] (csharpName id `withType` objectType)
--    name id = case idType id of
--	       ty@(PrimType _) -> csharpName id `withType` ty
--	       _               -> csharpName id `withType` objectType


csharp_top_bind (bndr, _) = [] -- ToDo
\end{code}

%************************************************************************
%*									*
\subsection{Expressions}
%*									*
%************************************************************************

\begin{code}
csharpVar :: Id -> Expr
csharpVar v | isExternalName (idName v) = mkNew (csharpIdType v) []
	  | otherwise	  	    =   Var (csharpName v)

csharpLit :: Literal.Literal -> Expr
csharpLit (MachInt i)  = Literal (IntLit (fromInteger i))
csharpLit (MachChar c) = Literal (CharLit c)
csharpLit (MachStr fs) = Literal (StringLit str)
   where
	str = concatMap renderString (unpackFS fs) ++ "\\000"
	-- This should really handle all the chars 0..31.
	renderString '\NUL' = "\\000"
	renderString other  = [other]

csharpLit other	     = pprPanic "csharpLit" (ppr other)

-- Pass in the 'shape' of the result.
csharpStgExpr :: (Expr -> Statement) -> StgExpr -> [Statement]
-- Generate code to apply the value of
-- the expression to the arguments aleady on the stack
csharpStgExpr r (StgLit lit)   = [r (csharpLit lit)]
csharpStgExpr r (StgApp func args)   = [r (csharpApp func args)]
csharpStgExpr r (StgConApp con args)   = [r (csharpConApp con args)]
csharpStgExpr r (StgOpApp op args ty)   = [r (csharpOpApp op args ty)]
csharpStgExpr r (StgLam _ bndrs body)   = csharpLam r bndrs body
csharpStgExpr r e@(StgLet bind expr) = csharpBind bind ++ csharpStgExpr r expr
csharpStgExpr r e@(StgLetNoEscape lvs_whole lvs_rhss bind expr) = csharpBind bind ++ csharpStgExpr r expr -- o que  NoEscape ???
csharpStgExpr r (StgSCC cc expr) = csharpStgExpr r expr -- ???
csharpStgExpr r e@(StgCase expr lvs_whole lvs_rhss bndr srt alt_type alts)
     = csharpCase r expr bndr alts
csharpStgExpr r other   = pprPanic "csharpStgExpr" (ppr other)

stgExprToExpr :: StgExpr -> Expr
stgExprToExpr (StgLit lit)   = (csharpLit lit)
stgExprToExpr (StgApp v [])   = (csharpVar v)
stgExprToExpr (StgApp func args)   = csharpApp func args
stgExprToExpr (StgConApp con args)   = csharpConApp con args
stgExprToExpr (StgOpApp op args ty)   = csharpOpApp op args ty
-- stgExprToExpr (StgLam _ bndrs body)   = csharpLam r bndrs body
-- stgExprToExpr e@(StgLet bind expr)
--    = map exprStmtToExpr (csharpBind bind) ++ [stgExprToExpr expr]
-- stgExprToExpr e@(StgLetNoEscape lvs_whole lvs_rhss bind expr)
--   = map exprStmtToExpr (csharpBind bind) ++ stgExprToExpr expr
stgExprToExpr (StgSCC cc expr) = stgExprToExpr expr
-- stgExprToExpr e@(StgCase expr lvs_whole lvs_rhss bndr srt alt_type alts)
--      = csharpCase r expr bndr alts
stgExprToExpr other   = pprPanic "stgExprToExpr" (ppr other)

exprStmtToExpr :: Statement -> Expr
exprStmtToExpr (ExprStatement e)  = e
exprStmtToExpr other   = error ("exprStmtToExpr:" ++ (show other))


-----------------------------------------------------------------
-- Code generation for case
-----------------------------------------------------------------

-- 	case e of x { Nil      -> r1
--		      Cons p q -> r2 }
-- ==>
--	final Object x = VM.WHNF(...code for e...)
--	else if x instance_of Nil {
--		...translation of r1...
--	} else if x instance_of Cons {
--		final Object p = ((Cons) x).f1
--		final Object q = ((Cons) x).f2
--		...translation of r2...
--	} else throw csharp.lang.Exception

-- This first special case happens a lot, typically
-- during dictionary deconstruction.
-- We need to access at least *one* field, to check to see
-- if we have correct constructor.
-- If we've got the wrong one, this is _|_, and the
-- casting will catch this with an exception.

{-
csharpCase r e bndr [(DataAlt d,bs,use,rhs)] | notNull bs
  = csharp_expr PushExpr e ++
    [ var [Final] (csharpName x)
	          (whnf primRep (vmPOP (primRepToType primRep))) ] ++
    bind_args d bs ++
    csharpExpr r rhs
   where
     primRep = idPrimRep x
     whnf PtrRep = vmWHNF	-- needs evaluation
     whnf _      = id		-- anything else does notg

     bind_args d bs = [var [Final] (csharpName b)
			   (Access (Cast (csharpConstrWkrType d) (csharpVar x)
				   ) f
			   )
		      | (b,f) <- filter isId bs `zip` (constrToFields d)
		      , not (isDeadBinder b)
		      ]
-}


-- ToDo: maybe can avoid the assignment when e@(StgApp v [])?
-- ToDo: use switch for primitive types  -- qual  o papel de Id ???
csharpCase :: (Expr -> Statement) -> StgExpr -> Id -> [StgAlt] -> [Statement]
csharpCase r e bndr alts
  | isPrimAlt con_alts =
	      [ var [] scrutinee (stgExprToExpr e)
               , Switch (Var scrutinee) (map mk_alt con_alts) (Just default_code)
              ]

  | otherwise =
	      [ var [] scrutinee (stgExprToExpr e)
              , IfThenElse (map (fix_stmts . mk_alt) con_alts)
                           (Just (mk_stmt default_code))
              ]
  where
    isPrimAlt ((LitAlt _,_,_,_):_) = True
    isPrimAlt _ = False
    scrutinee = csharpName bndr
    (con_alts, maybe_default) = findStgDefault alts
    default_code = case maybe_default of
			Nothing  -> [ExprStatement (Raise excName [Literal (StringLit "case failure")])]
			Just rhs -> (csharpStgExpr r rhs)
    mk_alt (DataAlt d,  bs, use, rhs)
      = (instanceOf bndr d, (bind_args d bs ++ csharpStgExpr r rhs))
    mk_alt (LitAlt lit, _bs, use, rhs)
      = (eqLit lit, csharpStgExpr r rhs)
    mk_alt other = pprPanic "mk_alt" (ppr other)

    eqLit (MachInt n) = (Literal (IntLit n))
    eqLit (MachChar n) = (Literal (CharLit n))
    eqLit other       = pprPanic "eqLit" (ppr other)
    bind_args d bs -- ??? b  o construtor todo ou os campos/aegumentos?
      = [var [] (csharpName b)
                   (Access
                      (Cast (csharpConstrWkrType d) (csharpVar bndr))
                      f
       		   )
       	 | (b,f) <- filter isId bs `zip` (constrToFields d)
       	   , not (isDeadBinder b)
       	]
    fix_stmts (c, stmtlist) = (c, mk_stmt stmtlist)
    mk_stmt [stmt] = stmt
    mk_stmt stmtlist = Block stmtlist

{-
csharpCase r e bndr alts
  | isIfThenElse && isPrimCmp
  = csharpIfThenElse r (fromJust maybePrim) tExpr fExpr
  | otherwise
  = csharp_expr PushExpr e ++
       [ var [Final] (csharpName x)
		           (whnf primRep (vmPOP (primRepToType primRep)))
       , IfThenElse (map mk_alt con_alts) (Just default_code)
       ]
  where
     isIfThenElse = CoreUtils.exprType e `Type.eqType` boolTy
		    -- also need to check that x is not free in
		    -- any of the branches.
     maybePrim    = findCmpPrim e []
     isPrimCmp    = isJust maybePrim
     (_,_,tExpr)  = CoreUtils.findAlt (DataAlt trueDataCon) alts
     (_,_,fExpr)  = CoreUtils.findAlt (DataAlt falseDataCon) alts

     primRep = idPrimRep x
     whnf PtrRep = vmWHNF	-- needs evaluation
     whnf _      = id

     (con_alts, maybe_default) = CoreUtils.findDefault alts
     default_code = case maybe_default of
			Nothing  -> ExprStatement (Raise excName [Literal (StringLit "case failure")])
			Just rhs -> Block (csharpExpr r rhs)

     mk_alt (DataAlt d,  bs, rhs) = (instanceOf x d, Block (bind_args d bs ++ csharpExpr r rhs))
     mk_alt (LitAlt lit, bs, rhs) = (eqLit lit     , Block (csharpExpr r rhs))
--     mk_alt (DEFAULT,[] , rhs) = (eqLit lit     , Block (csharpExpr r rhs))


     eqLit (MachInt n) = Op (Literal (IntLit n))

			    "=="
			    (Var (csharpName x))
     eqLit (MachChar n) = Op (Literal (CharLit n))
			    "=="
			    (Var (csharpName x))
     eqLit other       = pprPanic "eqLit" (ppr other)

     bind_args d bs = [var [Final] (csharpName b)
			   (Access (Cast (csharpConstrWkrType d) (csharpVar x)
				   ) f
			   )
		      | (b,f) <- filter isId bs `zip` (constrToFields d)
		      , not (isDeadBinder b)
		      ]
-}
csharpCase r e bndr alts  = pprPanic "csharpCase" (ppr alts)

csharpIfThenElse :: (Expr -> Statement) -> Expr -> StgExpr -> StgExpr -> [Statement]
csharpIfThenElse r cmp tExpr fExpr
{-
 - Now what we need to do is generate code for the if/then/else.
 - [all arguments are already check for simpleness (Var or Lit).]
 -
 - if (<prim> arg1 arg2 arg3 ...) {
 -	trueCode
 -  } else {
 -	falseCode
 - }
 -}
 = [IfThenElse [(cmp,j_tExpr)] (Just j_fExpr)]
 where
   j_tExpr, j_fExpr :: Statement
   j_tExpr = Block (csharpStgExpr r tExpr)
   j_fExpr = Block (csharpStgExpr r fExpr)

csharpBind :: StgBinding -> [Statement]
csharpBind (StgNonRec x (StgRhsCon _cc con args))
  = [ExprStatement
        (Assign (Var name)
        (New (csharpConstrWkrType con) (csharpArgs args) Nothing))
    ]
  where
    name = csharpName x `withType` closureType

csharpBind (StgNonRec x (StgRhsClosure _ _ fvs upd _ args rhs))
{-
	x = ...rhs_x...
  ==>
	final Object x = new Thunk( new Code() { ...code for rhs_x... } )
-}

-- ToDo: handle args and fvs
  = [ExprStatement (Assign (Var name) (stgExprToExpr rhs))]
  where
    name = csharpName x `withType` closureType

{- 	rec { x = ...rhs_x...; y = ...rhs_y... }
  ==>
	class x implements Code {
	  Code x, y;
	  public Object ENTER() { ...code for rhs_x...}
	}
	...ditto for y...

	final x x_inst = new x();
	...ditto for y...

	final Thunk x = new Thunk( x_inst );
	...ditto for y...

	x_inst.x = x;
	x_inst.y = y;
	...ditto for y...
-}
csharpBind (StgRec prs)  -- e quando for recursivo ???
  = concat $ map (csharpBind . (\(b,r) -> StgNonRec b r)) prs
{-
csharpBind (StgRec prs)
  = (map mk_class prs) ++ (map mk_inst prs) ++
    (map mk_thunk prs) ++ concat (map mk_knot prs)
  where
    mk_class (b,r) = Declaration (Class [] class_name [] [codeName] stmts)
		   where
		     class_name = csharpIdTypeName b
		     stmts = [Field [] (csharpName b `withType` codeType) Nothing | (b,_) <- prs] ++
			     [Method [Public] enterName [vmArg] [excName] (csharpExpr vmRETURN r)]

    mk_inst (b,r) = var [Final] name (mkNew ty [])
	where
	   name@(Name _ ty)  = csharpInstName b

    mk_thunk (b,r) = var [Final] (csharpName b `withType` codeType)
			 (mkNew thunkType [Var (csharpInstName b)])

    mk_knot (b,_) = [ ExprStatement (Assign lhs rhs)
		    | (b',_) <- prs,
		      let lhs = Access (Var (csharpInstName b)) (csharpName b'),
		      let rhs = Var (csharpName b')
		    ]
-}
csharpBind _ = error "csharpBind"

csharpLam :: (Expr -> Statement) -> [Id] -> StgExpr -> [Statement]
csharpLam r  []  body  = csharpStgExpr r body
-- csharpLam r (bndrs, body)
--   =  vmCOLLECT (length val_bndrs) this
--   ++ [var [Final] n (vmPOP t) | n@(Name _ t) <- val_bndrs]
--   ++ csharpExpr r body

csharpConApp :: DataCon -> [StgArg] -> Expr
csharpConApp con as
	= (New (csharpConstrWkrType con) (csharpArgs as) Nothing)

csharpApp :: Id -> [StgArg] -> Expr
csharpApp f as = (Call (csharpVar f)
                       (Name "" valueType)
                       (csharpArgs as))

-- csharpOpApp :: StgOp -> [StgArg] -> TypeRep.Type -> Expr
csharpOpApp op [arg1, arg2] ty = Op (csharpArg arg1)
                                    (csharpOp op)
                                    (csharpArg arg2)
  where
    csharpOp (StgPrimOp CharGtOp) = ">"
    csharpOp (StgPrimOp CharGeOp) = ">="
    csharpOp (StgPrimOp CharEqOp) = "=="
    csharpOp (StgPrimOp CharNeOp) = "/="
    csharpOp (StgPrimOp CharLtOp) = "<"
    csharpOp (StgPrimOp CharLeOp) = "<="

    csharpOp (StgPrimOp IntGtOp) = ">"
    csharpOp (StgPrimOp IntGeOp) = ">="
    csharpOp (StgPrimOp IntEqOp) = "=="
    csharpOp (StgPrimOp IntNeOp) = "/="
    csharpOp (StgPrimOp IntLtOp) = "<"
    csharpOp (StgPrimOp IntLeOp) = "<="

    csharpOp (StgPrimOp IntAddOp) = "+"
    csharpOp (StgPrimOp IntSubOp) = "-"
    csharpOp (StgPrimOp IntMulOp) = "*"
    csharpOp (StgPrimOp IntQuotOp) = "/"

    csharpOp (StgPrimOp primop) = pprPanic "csharpOp StgPrimOp" (ppr primop)
    csharpOp (StgFCallOp fcall _) = pprPanic "csharpOp StgFcall" (ppr fcall)
		

csharpArgs :: [StgArg] -> [Expr]
csharpArgs as = map csharpArg as 

csharpArg :: StgArg -> Expr
csharpArg (StgVarArg id) =  csharpVar id 
csharpArg (StgLitArg lit) = csharpLit lit 
csharpArg (StgTypeArg ty) = error "csharp_arg StgTypeArg"

-- The result is a list of statments that have the effect of
-- pushing onto the stack (via one of the VM.PUSH* commands)
-- the argument, (or returning, or setting a variable)
-- perhaps thunked.

{- This is mixing two things.
 (1) Optimizations for things like primitives, whnf calls, etc.
 (2) If something needs a thunk constructor round it.
 - Seperate them at some point!
 -}
data ExprRetStyle = SetVar Name | PushExpr | ReturnExpr

-- csharp_expr :: ExprRetStyle -> StgExpr -> [Statement]
-- csharp_expr _ (CoreSyn.Type t) = pprPanic "csharp_expr" (ppr t)
{-
csharp_expr ret e
   | isPrimCall = [push (fromJust maybePrim)]
	-- This is a shortcut, 
	-- basic names and literals do not need a code block
	-- to compute the value.
   | isPrim primty && CoreUtils.exprIsTrivial e = csharpExpr push e
   | isPrim primty =
 	  let expr  = csharpExpr vmRETURN e
	      code  = access (vmWHNF (newCode expr)) (primRepToType primty)
	  in [push code]
   | otherwise =
 	  let expr  = csharpExpr vmRETURN e
	      code  = newCode expr
	      code' = if CoreUtils.exprIsValue e 
		      || CoreUtils.exprIsTrivial e 
		      || isPrim primty
		      then code
		      else newThunk code
	  in [push code']
   where
	maybePrim  = findFnPrim e []
	isPrimCall = isJust maybePrim

	push e = case ret of
		  SetVar name -> var [Final] name e
		  PushExpr -> vmPUSH e
		  ReturnExpr -> vmRETURN e
	corety = CoreUtils.exprType e
	primty = Type.typePrimRep corety
	isPrim PtrRep  = False	-- only this needs updated
	isPrim _       = True

-}

coreTypeToType = primRepToType . Type.typePrimRep . CoreUtils.exprType

renameForKeywords :: (NamedThing name) => name -> String
renameForKeywords name 
  | str `elem` keywords = "zdk" ++ str
  | otherwise            = str
  where
	str = getOccString name

keywords :: [String]
keywords =
	[ "return"
	, "if"
	, "then"
	, "else"
	, "class"
	, "instance"
	, "import"
	, "throw"
	, "try"
	]

\end{code}

%************************************************************************
%*									*
\subsection{Helper functions}
%*									*
%************************************************************************

\begin{code}
true, this,csharpNull :: Expr
this = Var thisName 
true = Var (Name "true" (PrimType PrimBoolean))
csharpNull = Var (Name "null" objectType)

vmCOLLECT :: Int -> Expr -> [Statement]
vmCOLLECT 0 e = []
vmCOLLECT n e = [ExprStatement 
		    (Call varVM collectName
			[ Literal (IntLit (toInteger n))
			, e
			]
		    )
		]

vmPOP :: Type -> Expr 
vmPOP ty = Call varVM (Name ("POP" ++ suffix ty) ty) []

vmPUSH :: Expr -> Statement
vmPUSH e = ExprStatement 
	     (Call varVM (Name ("PUSH" ++ suffix (exprType e)) void) [e])

vmRETURN :: Expr -> Statement
vmRETURN e = Return (
     case ty of
	PrimType _ -> Call varVM (Name ("RETURN" ++ suffix ty)
				       valueType
				 ) [e]
	_ -> e)
  where
	ty = exprType e

var :: [Modifier] -> Name -> Expr -> Statement
var ms field_name@(Name _ ty) value 
   | exprType value == ty = Declaration (Field ms field_name (Just value))
   | otherwise            = var ms field_name (Cast ty value)

vmWHNF :: Expr -> Expr
vmWHNF e = Call varVM whnfName [e]

suffix :: Type -> String
suffix (PrimType t) = primName t
suffix _            = ""

primName :: PrimType -> String
primName PrimInt       = "int"
primName PrimChar      = "char"
primName PrimByte      = "byte"
primName PrimBoolean   = "boolean"
primName _             = error "unsupported primitive"

varVM :: Expr
varVM = Var vmName 

instanceOf :: Id -> DataCon -> Expr
instanceOf x data_con
  = InstanceOf (Var (csharpName x)) (csharpConstrWkrType data_con)

newCode :: [Statement] -> Expr
newCode [Return e] = e
newCode stmts	   = New codeType [] (Just [Method [Public] enterName [vmArg] [excName] stmts])

newThunk :: Expr -> Expr
newThunk e = New thunkType [e] Nothing

vmArg :: Parameter
vmArg = Parameter [Final] vmName

-- This is called with boolean compares, checking 
-- to see if we can do an obvious shortcut.
-- If there is, we return a (GOO) expression for doing this,

-- So if, we have case (#< x y) of { True -> e1; False -> e2 },
-- we will call findCmpFn with (#< x y), this return Just (Op x "<" y)

findCmpPrim :: CoreExpr -> [Expr] -> Maybe Expr
findCmpPrim (CoreSyn.App f a) as =
     case a of
	CoreSyn.Var v -> findCmpPrim f (csharpVar v:as)
	CoreSyn.Lit l -> findCmpPrim f (csharpLit l:as)
	_ -> Nothing
findCmpPrim (CoreSyn.Var p)   as = 
	case isPrimOpId_maybe p of
	  Just prim -> find_cmp_prim prim as
	  Nothing   -> Nothing
findCmpPrim _                 as = Nothing

find_cmp_prim cmpPrim args@[a,b] = 
   case cmpPrim of
     IntGtOp -> fn ">"
     IntGeOp -> fn ">="
     IntEqOp -> fn "=="
     IntNeOp -> fn "/="
     IntLtOp -> fn "<"
     IntLeOp -> fn "<="
     _ -> Nothing
  where
	fn op = Just (Op a op b)
find_cmp_prim _ _ = Nothing

findFnPrim :: CoreExpr -> [Expr] -> Maybe Expr
findFnPrim (CoreSyn.App f a) as =
     case a of
	CoreSyn.Var v -> findFnPrim f (csharpVar v:as)
	CoreSyn.Lit l -> findFnPrim f (csharpLit l:as)
	_ -> Nothing
findFnPrim (CoreSyn.Var p)   as = 
	case isPrimOpId_maybe p of
	  Just prim -> find_fn_prim prim as
	  Nothing   -> Nothing
findFnPrim _                 as = Nothing

find_fn_prim cmpPrim args@[a,b] = 
   case cmpPrim of
     IntAddOp -> fn "+"
     IntSubOp -> fn "-"
     IntMulOp -> fn "*"
     _ -> Nothing
  where
	fn op = Just (Op a op b)
find_fn_prim _ _ = Nothing
\end{code}

%************************************************************************
%*									*
\subsection{Haskell to CSharp Types}
%*									*
%************************************************************************

\begin{code}
exprType (Var (Name _ t)) = t
exprType (Literal lit)    = litType lit
exprType (Cast t _)       = t
exprType (New t _ _)      = t
exprType (Call _ (Name _ t) _) = t
exprType (Access _ (Name _ t)) = t
exprType (Raise t _)           = error "do not know the type of raise!"
exprType (Op _ op _) | op `elem` ["==","/=","<","<=","=>",">"]
		     = PrimType PrimBoolean
exprType (Op x op _) | op `elem` ["+","-","*"]
		     = exprType x
exprType expr = error ("can't figure out an expression type: " ++ show expr)

litType (IntLit i)    = PrimType PrimInt
litType (CharLit i)   = PrimType PrimChar
litType (StringLit i) = stringType	-- later, might use char array?
\end{code}

%************************************************************************
%*									*
\subsection{Name mangling}
%*									*
%************************************************************************

\begin{code}
codeName, excName, thunkName, closureName :: TypeName
codeName  = "haskell.runtime.Code"
thunkName = "haskell.runtime.Thunk"
closureName = "Closure"
excName   = "csharp.lang.Exception"

enterName, vmName,thisName,collectName, whnfName :: Name
enterName   = Name "ENTER"   objectType
vmName      = Name "VM"      vmType
thisName    = Name "this"    (Type "<this>")
collectName = Name "COLLECT" void
whnfName    = Name "WHNF"    objectType

fieldName :: Int -> Type -> Name	-- Names for fields of a constructor
fieldName n ty = Name ("f" ++ show n) ty

withType :: Name -> Type -> Name
withType (Name n _) t = Name n t

-- This maps (local only) names Ids to Names, 
-- using the same string as the Id.
csharpName :: Id -> Name
csharpName n 
--  | isExternalName (idName n) = error "using csharpName on global"
  | isExternalName (idName n) = Name (getOccString n)
		                     (primRepToType (idPrimRep n))
  | otherwise = Name (getOccString n)
		     (primRepToType (idPrimRep n))

-- TypeName's are almost always global. This would typically return something
-- like Test.foo or Test.Foozdc or PrelBase.foldr.
-- Local might use locally bound types, (which do not have '.' in them).

csharpIdTypeName :: Id -> TypeName
csharpIdTypeName n
    | isInternalName n' = renameForKeywords n'
    | otherwise      = moduleString (nameModule n') ++ "." ++ renameForKeywords n'
  where
	     n' = getName n

-- There is no such thing as a local type constructor.

csharpTyConTypeName :: TyCon -> TypeName
csharpTyConTypeName n = (moduleString (nameModule n') ++ "." ++ renameForKeywords n')
  where
	     n' = getName n

-- this is used for getting the name of a class when defining it.
shortName :: TypeName -> TypeName
shortName = reverse . takeWhile (/= '.') . reverse

-- The function that makes the constructor name
-- The constructor "Foo ..." in module Test,
-- would return the name "Test.Foo".

csharpConstrWkrName :: DataCon -> TypeName
csharpConstrWkrName = csharpIdTypeName . dataConWorkId

-- Makes x_inst for Rec decls
-- They are *never* is primitive
-- and always have local (type) names.
csharpInstName :: Id -> Name
csharpInstName n = Name (renameForKeywords n ++ "zdi_inst")
		      (Type (renameForKeywords n))
\end{code}

%************************************************************************
%*									*
\subsection{Types and type mangling}
%*									*
%************************************************************************

\begin{code}
-- Haskell RTS types
codeType, thunkType, valueType :: Type
codeType   = Type codeName
thunkType  = Type thunkName
valueType  = Type "haskell.runtime.Value"
vmType     = Type "haskell.runtime.VMEngine"

-- Basic CSharp types
closureType, objectType, stringType :: Type
closureType = Type "Closure"
objectType = Type "object"
stringType = Type "string"

void :: Type
void = PrimType PrimVoid

inttype :: Type
inttype = PrimType PrimInt

chartype :: Type
chartype = PrimType PrimChar

bytetype :: Type
bytetype = PrimType PrimByte

-- This lets you get inside a possible "Value" type,
-- to access the internal unboxed object.
access :: Expr -> Type -> Expr
access expr (PrimType prim) = accessPrim (Cast valueType expr) prim
access expr other           = expr

accessPrim expr PrimInt  = Call expr (Name "intValue" inttype) []
accessPrim expr PrimChar = Call expr (Name "charValue" chartype) []
accessPrim expr PrimByte = Call expr (Name "byteValue" bytetype) []
accessPrim expr other    = pprPanic "accessPrim" (text (show other))

-- This is where we map from typename to types,
-- allowing to match possible primitive types.
mkType :: TypeName -> Type
mkType "PrelGHC.Intzh"  = inttype
mkType "PrelGHC.Charzh" = chartype
mkType other            = Type other

-- Turns a (global) Id into a Type (fully qualified name).
csharpIdType :: Id -> Type
csharpIdType = mkType . csharpIdTypeName

csharpLocalIdType :: Id -> Type
csharpLocalIdType = primRepToType . idPrimRep

primRepToType ::PrimRep -> Type
primRepToType PtrRep  = objectType
primRepToType IntRep  = inttype
primRepToType CharRep = chartype
primRepToType Int8Rep = bytetype
primRepToType AddrRep = objectType
primRepToType other   = pprPanic "primRepToType" (ppr other)

-- The function that makes the constructor name
csharpConstrWkrType :: DataCon -> Type
csharpConstrWkrType con = Type (csharpConstrWkrName con)
\end{code}

